/* -*- mode: c++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

/*
 Copyright (C) 2011 Klaus Spanderen

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the license for more details.
*/

/*! \file gpubicgstab.cu
*/

#if !defined(QL_NO_UBLAS_SUPPORT)
#include <ql/experimental/cuda/cudatypes.hpp>
#include <ql/experimental/cuda/gpubicgstab.hpp>

#include <cusp/csr_matrix.h>
#include <cusp/krylov/bicgstab.h>

#include <vector>
#include <algorithm>

namespace QuantLib {

    class GPUBiCGStab::pimpl {
      public:
        typedef cusp::csr_matrix<int32_t, real, cusp::device_memory>
            gpu_matrix_type;

        typedef gpu_matrix_type::index_type   index_type;
        typedef gpu_matrix_type::value_type   value_type;
        typedef gpu_matrix_type::memory_space memory_space;

        typedef cusp::array1d<value_type, memory_space> array1d_type;

        boost::scoped_ptr<gpu_matrix_type> A_;
    };

    GPUBiCGStab::GPUBiCGStab(const matrix_type& m)
    : pImpl_(new pimpl) {

        cusp::csr_matrix<pimpl::index_type, pimpl::value_type,
                         cusp::host_memory> h_A(m.size1(), m.size2(),
                                                m.value_data().size());

        h_A.row_offsets    = m.index1_data();
        h_A.column_indices = m.index2_data();
        h_A.values         = m.value_data();

        pImpl_->A_.reset(new pimpl::gpu_matrix_type(h_A));
    }

    GPUBiCGStab::~GPUBiCGStab() {
        // declare destructor here so that the compiler knows pimpl
    }

    Disposable<Array> GPUBiCGStab::solve(const Array& h_b,
                                         Size maxIter, Real eps) {

        const Size n = h_b.size();

        std::vector<pimpl::value_type> hv_b(n);
        std::transform(h_b.begin(), h_b.end(), hv_b.begin(),
                       conversion<Array::value_type, pimpl::value_type>());

        pimpl::array1d_type b = hv_b;
        pimpl::array1d_type x(n, 0.0);

        cusp::default_monitor<pimpl::value_type> monitor(b, maxIter, eps);
        cusp::krylov::bicgstab(*pImpl_->A_, x, b, monitor);

        const cusp::array1d<pimpl::value_type, cusp::host_memory> hv_x = x;

        Array retVal(n);
        std::transform(hv_x.begin(), hv_x.end(), retVal.begin(),
                       conversion<pimpl::value_type, Array::value_type>());

        return retVal;
    }
}

#endif
